#!/usr/bin/env python3

import matplotlib.pyplot as plt
import numpy as np
import subprocess
import os

RESULT_FILENAME = "result.png"
RESULT_SPEEDUP_FILENAME = "result_speedup.png"
RESULT_APPEND_FILENAME = "result_and_speedup.png"


def plot(xs, ys, stds):
    fig, ax = plt.subplots()
    ax.plot(xs, ys, label="OpenCL")
    ax.set(xlabel="Nombre de tâches", ylabel='Temps (secondes)',
           title=f"Jeu de la vie\nCalcul de 100 générations avec \nune matrice de 30'000x30'000")
    ax.grid()
    plt.tight_layout()
    plt.legend()
    fig.savefig(RESULT_FILENAME)
    plt.show()


def plot_speedup(xs, ys):
    fig, ax = plt.subplots()
    ax.plot(xs, ys, label="Speedup obtenu")
    ax.plot(xs, xs, label="Speedup idéal")
    ax.set(xlabel="Nombre de tâches", ylabel='Speedup',
           title=f"Jeu de la vie\nSpeedup du calcul de 100 générations avec \nune matrice de 30'000x30'000")
    ax.grid()
    plt.tight_layout()
    plt.legend()
    fig.savefig(RESULT_SPEEDUP_FILENAME)
    plt.show()


def main():
    results = np.genfromtxt('results/gol-opencl-2021-06-13.csv', delimiter=';')
    results_ordered = {}
    for result in results:
        key = int(result[0])
        if key in results_ordered:
            results_ordered[key].append(result[-1])
        else:
            results_ordered[int(result[0])] = [result[-1]]
    xs = []
    ys = []
    stds = []
    speedups = []
    for key in results_ordered.keys():
        xs.append(key)
        ys.append(sum(results_ordered[key]) / len(results_ordered[key]))
        stds.append(np.std(results_ordered[key]))
        speedups.append(ys[0] / ys[-1])
    plot(xs, ys, stds)
    plot_speedup(xs, speedups)
    print("Average")
    for y in ys:
        print(round(y, 3))
    print()
    print("Stds")
    for std in stds:
        print(round(std, 3))
    # Need ImageMagick
    subprocess.run(["docker", "run", "-v", f"{os.getcwd()}:/imgs", "dpokidov/imagemagick", "+append",
                    f"/imgs/{RESULT_FILENAME}", f"/imgs/{RESULT_SPEEDUP_FILENAME}", f"/imgs/{RESULT_APPEND_FILENAME}"],
                   capture_output=False)


if __name__ == '__main__':
    main()
